#include <io.h>
#include <fcntl.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <windows.h>
#endif

#include "zlib/zlib.h"

class conv
{
	template <class T> class xbuf
	{
		T *ptr;
		int len;
	public:
		xbuf()
		{
			ptr = 0;
		}
		~xbuf()
		{
			free();
		}
		void free()
		{
			if (ptr)
			{
				delete [] ptr;
				ptr = 0;
			}
			len = 0;
		}
		bool alloc(int s)
		{
//			printf("%d\n", s);
			free();
			ptr = new T [s];
			len = s;
			return true;
		}
		T *top()
		{
			return ptr;
		}
		unsigned dword(int ofs)
		{
			unsigned char *pb = (unsigned char *)ptr;
			unsigned ret = 0;
			ret |= ((unsigned)pb[ofs + 0]) << (8 * 0);
			ret |= ((unsigned)pb[ofs + 1]) << (8 * 1);
			ret |= ((unsigned)pb[ofs + 2]) << (8 * 2);
			ret |= ((unsigned)pb[ofs + 3]) << (8 * 3);
			return ret;
		}
		int size()
		{
			return len;
		}
	};
	class xfile
	{
		FILE *fp;
		int len;
	public:
		xfile()
		{
			fp = NULL;
		}
		void close()
		{
			if (fp)
			{
				fclose(fp);
				fp = NULL;
			}
			len = 0;
		}
		bool load(char *path)
		{
#ifdef _WIN32
			fp = fopen(path, "rb");
			len = GetFileSize((HANDLE)_get_osfhandle(_fileno(fp)), NULL);
#else
			fp = fopen(path, "r");
			fseek(fp, 0, SEEK_END);
			len = ftell(fp);
			fseek(fp, 0, SEEK_SET);
#endif
			return fp != NULL;
		}
		bool save(char *path)
		{
#ifdef _WIN32
			fp = fopen(path, "wb");
#else
			fp = fopen(path, "w");
#endif
			len = 0;
			return fp != NULL;
		}	
		int size()
		{
			return len;
		}
		bool read(void *buf, int bytes)
		{
			return fp && fread(buf, 1, bytes, fp) == bytes;
		}
		bool write(void *buf, int bytes)
		{
			return fp && fwrite(buf, 1, bytes, fp) == bytes;
		}
	};
	static void setword(unsigned char *p, long dw)
	{
		p[0] = (unsigned char)((dw >> (8 * 0)) & 0xff);
		p[1] = (unsigned char)((dw >> (8 * 1)) & 0xff);
	}
	static void setdword(unsigned char *p, long dw)
	{
		p[0] = (unsigned char)((dw >> (8 * 0)) & 0xff);
		p[1] = (unsigned char)((dw >> (8 * 1)) & 0xff);
		p[2] = (unsigned char)((dw >> (8 * 2)) & 0xff);
		p[3] = (unsigned char)((dw >> (8 * 3)) & 0xff);
	}
public:
	conv()
	{
	}
	~conv()
	{
	}
	int errmsg(int n, char *sarg = 0, int larg = 0)
	{
		switch (n)
		{
		case 1:
			fprintf(stderr, "%s\n", "Short of memory");
			break;
		case 2:
			fprintf(stderr, "%s:%s\n", "File read error", sarg);
			break;
		case 3:
			fprintf(stderr, "%s:%d\n", "Broken file", larg);
			break;
		case 4:
			fprintf(stderr, "%s:%s\n", "File write error", sarg);
			break;
		}
		return n;
	}
	int run(char *path, int skip)
	{
		xfile lfile;
		xbuf<unsigned char> lbuf;


		if (!lfile.load(path)) return errmsg(2, path);
		if (!lbuf.alloc(lfile.size())) return errmsg(1);
		if (!lfile.read(lbuf.top(), lbuf.size())) return errmsg(2, path);

		lfile.close();

		if (lbuf.dword(0) & 0xffffff != 0x465350) return errmsg(3);
		
		unsigned res_size = lbuf.dword(4);
		unsigned prg_size = lbuf.dword(8);
		unsigned buf_size = lbuf.size() * 2;

		xbuf<unsigned char> sbuf;
		uLongf zsize;
		int zresult;
		do
		{
			buf_size += buf_size;
			if (!sbuf.alloc(buf_size)) return errmsg(1);
			zsize = sbuf.size();
			zresult = uncompress(sbuf.top(), &zsize, lbuf.top() + 16 + res_size, prg_size);
		}
		while (zresult == Z_BUF_ERROR);
		if (zresult < Z_OK) return errmsg(3, 0, zresult);

		lbuf.free();

		if (zsize > 0 && skip < zsize)
		{
			xbuf<char> nbuf;
			if (!nbuf.alloc(strlen(path) + 16)) return errmsg(1);
			strcpy(nbuf.top(), path);
			strcat(nbuf.top(), ".bin");

			xfile sfile;
			if (!sfile.save(nbuf.top())) return errmsg(4, nbuf.top());
			if (!sfile.write(sbuf.top() + skip, zsize - skip)) return errmsg(4, nbuf.top());
		}

		return 0;
	}
};

int main(int argc, char **argv)
{
	unsigned s = 0;
	for (int i = 1; i < argc; i++)
	{
		if (strcmpi(argv[i], "-s") == 0 && i + 1 < argc)
		{
			s = atoi(argv[++i]);
		}
		else
		{
			conv c;
			c.run(argv[i], s);
		}
	}
	return 0;
}

